#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import os

import time

import shutil

import numpy as np

from scipy import sparse

from scipy.sparse import linalg

from mayavi.api import OffScreenEngine

from pyface.image_resource import ImageResource

from traits.api import HasTraits, Instance, Property, Enum, Button,String,\
                       on_trait_change, Float, Bool, Int, Constant, ReadOnly,\
                       List, Array, Range

from traitsui.api import View, Item, HSplit, VSplit, InstanceEditor, HGroup,\
                         Group, ListEditor, Tabbed, VGroup, CheckListEditor,\
                         ArrayEditor, RangeEditor, Controller

#------------------------------------------------------------------------------

'''-'''

__author__ = 'Davide Pellis'


#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#                            MeshMovieManager
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

class M_Handler(Controller):

    def close(self, info, is_ok):
        info.object._closed = True
        Controller.close(self ,info, is_ok)
        return True


class MovieManager(HasTraits):

    #--------------------------------------------------------------------------
    #                                 Traits
    #--------------------------------------------------------------------------
    
    _handler = M_Handler()
    
    scene = String()
    
    scenes = []

    record = Bool(False)

    shoot_view_button = Button(label='Shoot view')

    shoot_frame = Button()

    make_movie_button = Button(label='Make movie')

    clear_button = Button(label='Clear')

    height = Int(800)

    width = Int(800)

    fps = Int(20)

    speed = Float(1)

    intro = Float(1, label='Intro [s]')

    final = Float(2, label='Final [s]')

    title = String('movie')

    key_frame = Int(0)

    K = Int(0)

    time = String('0.00', label='Time [s]')

    duration = Float(6, label='Rotation [s]')
    
    shoot_duration = Float(0.5, label='Duration [s]')

    pre_time = Float(0.5, label='Pre [s]')

    post_time = Float(0, label='Post [s]')

    camera_rotation = Button(label='Rotate')

    angle = Int(360, label='Angle [deg]')
    
    smoothing = Float(0.9)

    #--------------------------------------------------------------------------
    #                                  View
    #--------------------------------------------------------------------------

    view = View(
                VGroup(HGroup(Item('scene',
                                   resizable=True,
                                   editor=CheckListEditor(values=scenes),),
                              Item('record', show_label=True),
                              show_border=True),
                       HGroup(
                              VGroup(
                                     Item('fps',resizable=True,),
                                     Item('height',resizable=True,),
                                     Item('intro',resizable=True,),

                                     ),
                              VGroup(
                                     Item('speed',resizable=True,),
                                     Item('width',resizable=True,),
                                     Item('final',resizable=True,),
                                     ),
                              show_border=True),
                       
                       VGroup(
                              Item('key_frame',
                                   editor = RangeEditor(low=0,
                                                        high_name='K',
                                                        mode='slider'),
                                   ),
                              Item('time',
                                   style='readonly'),
                              show_border=True),
                       
                       VGroup(HGroup(Item('camera_rotation',show_label=False),
                                     Item('angle'),
                                     ),
                              HGroup(Item('pre_time'),
                                     Item('duration'),
                                     Item('post_time'),
                                     ),
                              show_border=True,),
                       VGroup(
                              Item('smoothing',
                                   editor = RangeEditor(low=0.,
                                                        high=1.,
                                                        mode='slider'),
                                   ),
                              show_border=True),
                       HGroup(
                              Item('shoot_frame', show_label=False),
                              Item('shoot_duration'),
                              show_border=True,),
                       HGroup(
                              Item('clear_button'),
                              Item('make_movie_button'),                             
                              show_border=True,
                              show_labels=False),
                     show_border=True,),
                resizable=False,
                handler=_handler,
                width=0.25,
                icon = ImageResource('icons/moviemanager.png'),
                title = 'Movie studio',
                )

    #--------------------------------------------------------------------------
    #                             Initialization
    #--------------------------------------------------------------------------

    def __init__(self):
        HasTraits.__init__(self)

        self._key_times =  np.array([])

        self.key_views = {}

        self._camera_rotations = {}

        self.t0 = 0

        self.state = 'start'

        self._R = 0

        self._N = 0

        self._scene_manager = None
        
        self._closed = True
        
        self._geolab = None
        
        self.__shoting = False
        
        self.__clearing = False
        
        self.__vertical_view = False

    #--------------------------------------------------------------------------
    #                            Set attributes
    #--------------------------------------------------------------------------

    @property
    def scene_manager(self):
        return self.geolab.scene_managers[self.scene]
    
    @property
    def geolab(self):
        return self._geolab
    
    @geolab.setter
    def geolab(self, geolab):
        self._geolab = geolab
        self.scenes.clear()
        self.scenes.extend(self.geolab.scene_keys)
        self.scene = self.scenes[0]
        self.clear()
        
    @property
    def total_frames(self):
        intro = int(self.intro * self.fps)
        final = int(self.final * self.fps)
        T = self._R + self._N + intro + final
        return T
    
    @on_trait_change('scene')
    def scene_changed(self):
        self.clear()
        
    @on_trait_change('clear_button')
    def clear_button_fired(self):
        self.clear()
        
    @on_trait_change('shoot_frame')
    def shoot_frame_fired(self):
        self.shoot(time_step=self.shoot_duration, kink=True)
        
    @on_trait_change('shoot_view_button')
    def shoot_view(self):
        view = self.scene_manager.get_position()
        self.key_views[str(self.key_frame)] = view
        
    @on_trait_change('record')
    def record_fired(self):
        pass#self.shoot()
    
    @on_trait_change('key_frame')
    def plot_key_frame(self):
        if not self.__shooting and not self.__clearing:
            self.scene_manager.plot_key_frame(self.key_frame)
            self.time = ('{:.2f}').format(self._key_times[self.key_frame])
    
    #--------------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def start(self):
        if not self._closed:
            self._handler.info.ui.dispose()
        self._closed = False
        self.configure_traits()

    def close(self):
        try:
            self._closed = True
            self._handler.info.ui.dispose()
        except:
            pass

    #--------------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
    
    def _plot_data(self, data):
        from geometrylab.vtkplot import Polyline as Poly
        from geometrylab.vtkplot import view
        from geometrylab.geometry import Polyline
        N = len(data)
        plot = []
        for i in range(data.shape[1]):
            Pt = np.column_stack((data[:,i,0], np.arange(N), np.zeros(N)))
            pl = Polyline(Pt)
            plP = Poly(pl, color='r')
            plot.append(plP)
        view(plot)
        
    #--------------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------
        
    def clear(self):
        self.__clearing = True
        self._key_times =  np.array([])
        self.key_views = {}
        self._camera_rotations = {}
        self.t0 = 0
        self.state = 'start'
        self._R = 0
        self._N = 0                    
        self.__shoting = False        
        self.__vertical_view = False
        self.K = 0
        self.key_frame = 0
        self.time = '0.00'
        self.scene_manager.clear_records()
        self.__clearing = False

    def shoot(self, time_step=0.1, kink=False):
        if self.record:
            self.__shooting = True
            '''
            t = time.time()       
            if self.state == 'start':
                self.t0 = t
                self.state = 'record'
            elif self.state == 'pause':
                self.t0 = t - self.t0
                self.state = 'record'
            if time_step is None:
                dt = np.array([t - self.t0])
            else:
                try:
                    dt = self._key_times[-1] + time_step
                except:
                    dt = 0
                self.pause()
            '''
            try:
                dt = self._key_times[-1] + time_step
            except:
                dt = 0
            self._key_times = np.hstack((self._key_times, dt))
            self.K = int(len(self._key_times) - 1)         
            self.scene_manager.shoot(kink)
            self.key_frame = self.K
            self.time = ('{:.2f}').format(self._key_times[-1])
            self.__shooting = False

    def pause(self, time_step=0.1):
        if self.state == 'pause':
            return
        self.state = 'pause'
        t = time.time()
        self.t0 = t - self.t0 + time_step

    def unpause(self):
        if self.state == 'record':
            return
        self.state = 'record'
        t = time.time()
        self.t0 = t - self.t0

    #--------------------------------------------------------------------------
    #                                 Camera
    #--------------------------------------------------------------------------

    def correct_views(self):
        for key in self._camera_rotations:
            if key in self.key_views:
                self.key_views[key][3:6] = np.array([0,0,1])

    #--------------------------------------------------------------------------
    #                             Plot functions
    #--------------------------------------------------------------------------
    
    def _plot_frame(self, frame_number):
        self.scene_manager.plot_frame(frame_number)
        self.save_image()

    #--------------------------------------------------------------------------
    #                            Interpolation
    #--------------------------------------------------------------------------
    
    def _interpolate(self):
        self._interpolate_times()
        self.scene_manager.interpolate_frames(self._N,self._key_frames_number)
    
    def _interpolate_times(self):
        duration = self._key_times[-1]
        N = int(duration * (self.fps/self.speed)) + 1
        key = np.array(self._key_times) * (self.fps/self.speed)
        key = np.array(key, dtype=np.int)
        self._key_frames_number = key
        self._N = max(N, len(key))
        self._key_frames_number[-1] = self._N - 1

    #--------------------------------------------------------------------------
    #                                View
    #--------------------------------------------------------------------------
    
    @on_trait_change('camera_rotation')
    def camera_rotation_fired(self):
        rotation = (self.angle, self.duration, self.pre_time, self.post_time)
        self._camera_rotations[str(self.key_frame)] = rotation
        time = (self.duration + self.pre_time + self.post_time)
        self._R += int(time*self.fps / self.speed)
        self.__vertical_view = True
        
    def _smooth_rotation(self, degrees, n):
        phi = np.linspace(0, 1, n//2)
        phi = 1 / (1 + np.e**(-15*(phi-0.4)))
        phi = np.hstack((phi,phi[::-1]))
        phi = degrees * phi / np.sum(phi)
        if len(phi) < n:
            phi = np.hstack((phi,[phi[-1]]))
        return phi
       
    def _rotate_view(self, time, degrees):
        N = int(time * self.fps)
        phi = self._smooth_rotation(degrees, N)
        for j in range(N):
            self.scene_manager.camera_rotation(phi[j])
            self.scene_manager.render()
            self.save_image()

    def _update_view(self, frame_number):
        k = np.where(self._key_frames_number == frame_number)[0]
        for f in k:
            i = str(f)
            if i in self._camera_rotations:           
                degrees = self._camera_rotations[i][0]
                time = self._camera_rotations[i][1]
                pre_time = self._camera_rotations[i][2]
                post_time = self._camera_rotations[i][3]
                self._yield_frame(pre_time)
                self._rotate_view(time, degrees)
                self._yield_frame(post_time)
                
    def vertical_view(self):
        self.scene_manager.vertical_view()

    #--------------------------------------------------------------------------
    #                                Movie maker
    #--------------------------------------------------------------------------

    def _make_paths(self):
        c_path = os.getcwd()
        n = 0
        forbidden = True
        while forbidden:
            video_name = '{}_{}.mp4'.format(self.title, n)
            folder = '{}_{}_img'.format(self.title, n)
            out_path = os.path.join(c_path,folder)
            video_path = os.path.join(c_path,video_name)
            prefix = '{}_{}'.format(self.title, n)
            n += 1
            if not os.path.exists(out_path) and not os.path.exists(video_path):
                os.makedirs(out_path)
                forbidden = False
        self._ext = '.png'
        self._out_path = out_path
        self._prefix = prefix
        self._padding = len(str(self.total_frames))
        self._i = 0

    def save_image(self):
        self._i += 1
        zeros = '0'*(self._padding - len(str(self._i)))
        suffix = '{}_{}{}{}'.format(self._prefix, zeros, self._i, self._ext)
        name = os.path.join(self._out_path, suffix)
        self._last_image = name
        self.scene_manager.save(name , size=(self.width,self.height))

    def _yield_frame(self, time):
        N = int(time * self.fps)
        for i in range(N):
            self._i += 1
            zeros = '0'*(self._padding - len(str(self._i)))
            suffix = '{}_{}{}{}'.format(self._prefix, zeros,self._i,self._ext)
            name = os.path.join(self._out_path, suffix)
            shutil.copyfile(self._last_image, name)

    def _make_mp4(self):
        form = '{}_%0{}d{}'.format(self._prefix, self._padding, self._ext)
        ffmpeg_fname = os.path.join(self._out_path, form)
        #print(ffmpeg_fname)
        cm = 'ffmpeg -f image2 -r {} -i {} -vcodec mpeg4 -qscale:v 1 -y {}.mp4'
        cm = cm.format(self.fps, ffmpeg_fname, self._prefix)
        try:
            os.system(cm)
            shutil.rmtree(self._out_path)
        except:
            print('Video not created: install FFmpeg on your system!')

    @on_trait_change('make_movie_button')
    def make_movie(self):
        self.geolab.set_state(None)        
        if len(self._key_times) < 1:
            return
        if self.__vertical_view:
            self.vertical_view()
        self._interpolate()
        self._make_paths()
        self._plot_frame(0)
        self._yield_frame(self.intro)
        for i in range(self._N):
            self._plot_frame(i)
            self._update_view(i)
        self._yield_frame(self.final)
        self._make_mp4()
        print('*** Video created ***')
        self.geolab.update_plot()
        self.key_frame = self.K
