#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from scipy import sparse

# -----------------------------------------------------------------------------

from geolab import utilities

from geolab.optimization.guidedprojectionbase import GuidedProjectionBase

# -----------------------------------------------------------------------------

__author__ = 'Davide Pellis'

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

class CombNormals(GuidedProjectionBase):

    _polyline = None
    
    _untwist = False

    def __init__(self):
        GuidedProjectionBase.__init__(self)

        weights = {

        'geometric' : 1,
        
        'normals' : 10,

        }

        self.add_weights(weights)

    #--------------------------------------------------------------------------
    #
    #--------------------------------------------------------------------------

    @property
    def polyline(self):
        return self._polyline

    @polyline.setter
    def polyline(self, polyline):
        self._polyline = polyline
        self.initialization()

    #--------------------------------------------------------------------------
    #                               Initialization
    #--------------------------------------------------------------------------

    def on_initialize(self):
        tangents = self.polyline.vertex_tangents()
        self.add_value('tangents', np.copy(tangents))
        normals, cos = self.polyline.vertex_bisectors(return_cosines=True)
        self.add_value('normals_0', np.copy(normals))
        self.add_value('normals', np.copy(normals))
        self.add_value('binormals', np.cross(tangents, normals))
        w = np.zeros(len(cos))
        u = np.zeros(len(cos))
        w[cos > -0.9] = 1
        #u[np.argmax(cos)] = 1
        if not self.polyline.closed:
            w[0] = w[-1] = 0
        else:
            w[0] = w[-1] = 1
        self.add_value('local_w', np.copy(w))
        self.add_value('untwist_w', np.copy(u))

    def set_dimensions(self):
        self._N = 3*self.polyline.V

    def initialize_unknowns_vector(self):
        normals = self.get_value('normals')
        X = normals.flatten('F')
        self._X = X
        self._X0 = np.copy(X)

    def make_errors(self):
        pass

    def post_iteration_update(self):
        pass

    def on_reinitilize(self):
        pass

    #--------------------------------------------------------------------------
    #                                   Utilities
    #--------------------------------------------------------------------------

    def normals(self, untwist=False):
        N = np.reshape(self.X[0:3*self.polyline.V], (self.polyline.V,3),
                       order='F')
        N = utilities.normalize(N)
        vertices = self.polyline.vertices
        if False:#not self.polyline.closed:
            T0 = (vertices[1] - vertices[0])
            N0 = np.cross(np.cross(T0,N[1,:]), T0)
            norm = np.linalg.norm(N0)
            N[0,:] = N0/norm
            T1 = (vertices[-1] - vertices[-2])
            N1 = np.cross(-np.cross(T0,N[-2,:]), T1)
            norm = np.linalg.norm(N1)
            N[-1,:] = N1/norm
        if untwist:
            V0 = N[1:self.polyline.V]
            V1 = N[0:self.polyline.V-1]
            D = np.einsum('ij,ij->i', V0, V1)
            sign = 1
            for i in range(len(D)):
                N[i] *= sign
                if D[i] < -0.2:
                    sign *= -1
            N[-1] = N[-2]
        return N

    def plot_normals(self):
        from geolab import plotter
        N = plotter.Vectors(self.normals(), anchor=self.polyline.vertices)
        w = self.get_value('local_w')
        w = np.column_stack((w,w,w))
        N0 = plotter.Vectors(self.get_value('normals'),
                             anchor=self.polyline.vertices,
                             color='gray_50')
        pl = plotter.Polyline(self.polyline, color='r')
        plotter.view([N,N0,pl])
        
    # -------------------------------------------------------------------------
    #                          Geometric Constraints
    # -------------------------------------------------------------------------

    def unit_constraints(self):
        w = 1
        V = self.polyline.V
        v = np.arange(V)
        i = np.arange(V)
        i = np.hstack((i, i, i))
        j = np.hstack((v, V+v, 2*V+v))
        X = self.X
        data = 2 * np.hstack((X[v], X[V+v], X[2*V+v])) * w
        H = sparse.coo_matrix((data,(i,j)), shape=(V, self.N))
        r = ((X[v]**2 + X[V+v]**2 + X[2*V+v]**2) + 1) * w
        self.add_iterative_constraint(H, r, 'unit')

    def tangent_constraints(self):
        w = 1 #- self.get_value('local_w')
        V = self.polyline.V
        t = self.get_value('tangents')
        v = np.arange(V)
        i = np.arange(V)
        i = np.hstack((i, i, i))
        j = np.hstack((v, V+v, 2*V+v))
        data = np.hstack((t[:,0]*w, t[:,1]*w, t[:,2]*w))
        H = sparse.coo_matrix((data,(i,j)), shape=(V, self.N))
        r = np.zeros(V)
        self.add_constant_constraint(H, r, 'tangent')
        
    def normal_constraints(self):
        if self._untwist:
            w = self.get_value('untwist_w') * self.get_weight('normals')
        else:
            w = self.get_value('local_w') * self.get_weight('normals')
        V = self.polyline.V
        b = self.get_value('binormals')
        v = np.arange(V)
        i = np.arange(V)
        i = np.hstack((i, i, i))
        j = np.hstack((v, V+v, 2*V+v))
        data = np.hstack((b[:,0]*w, b[:,1]*w, b[:,2]*w))
        H = sparse.coo_matrix((data,(i,j)), shape=(V, self.N))
        r = np.zeros(V)
        self.add_constant_constraint(H, r, 'normal')

    def normals_fairness(self):
        w = 0.1 / self.polyline.V
        v0 = np.arange(self.polyline.V)
        v2 = np.roll(v0,1)
        v1 = np.roll(v0,-1)
        if not self.polyline.closed:
            v0 = v0[1:-2]
            v2 = v2[1:-2]
            v1 = v1[1:-2]
        W = len(v0)
        V = self.polyline.V
        one  = np.ones(W)
        i = np.arange(W)
        i = np.hstack((i,i,i))
        j = np.hstack((v0,v1,v2))
        data = np.hstack((-2*one, one, one)) * w
        i = np.hstack((i, W+i, 2*W+i))
        j = np.hstack((j, V+j, 2*V+j))
        data = np.hstack((data, data, data))
        K = sparse.coo_matrix((data,(i,j)), shape=(3*W, self.N))
        s = np.zeros(3*W)
        self.add_constant_fairness(K, s, 'fairness')
        
    def difference_fairness(self):
        w = 1 / self.polyline.V
        v0 = np.arange(self.polyline.V)
        v1 = np.roll(v0,1)
        if not self.polyline.closed:
            v0 = v0[1:]
            v1 = v1[1:]
        W = len(v0)
        V = self.polyline.V
        one  = np.ones(W)
        i = np.arange(W)
        i = np.hstack((i,i))
        j = np.hstack((v0,v1))
        data = np.hstack((-one, one)) * w
        i = np.hstack((i, W+i, 2*W+i))
        j = np.hstack((j, V+j, 2*V+j))
        data = np.hstack((data, data, data))
        K = sparse.coo_matrix((data,(i,j)), shape=(3*W, self.N))
        s = np.zeros(3*W)
        self.add_constant_fairness(K, s, 'fairness')

    # -------------------------------------------------------------------------
    #                                 Build
    # -------------------------------------------------------------------------

    def build_iterative_constraints(self):
        self.unit_constraints()
        self.tangent_constraints()
        self.normal_constraints()

    def build_constant_fairness(self):
        self.difference_fairness()

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------

def comb_normals(polyline):
    optimizer = CombNormals()
    optimizer.verbose = False
    optimizer.polyline = polyline
    optimizer.threshold = 1e-3
    optimizer.iterations = 10
    optimizer.epsilon = 0.0001
    optimizer.optimize()
    #optimizer.plot_normals()
    return optimizer.normals(untwist=True)
