#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from scipy import sparse

from scipy.sparse import linalg

#------------------------------------------------------------------------------


'''-'''

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                               Mesh Recorder
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class MeshRecorder(object):

    def __init__(self, mesh_manager=None):
        
        self._key_frames_number = None
        
        self._frame_number = 0
        
        self._key_number = 0
        
        self._N = None
        
        self.__recording = False
        
        self.__shooting = False
        
        self.__key = False
        
        self._mesh_manager = mesh_manager
        
        self._vertices = None

        self._faces = {}

        self._edges = {}

        self._points = {}

        self._selected_faces = []

        self._selected_edges = []
        
        self._selected_vertices = []
        
        self._glyphs = {}
        
        self._interpolated_data = {}
        
        self._sources_on_scene = {}
        
        self._key_register = []
        
        self._kinks = np.array([], dtype=np.bool)
        
        self.__indices_updated = False
        
    # -------------------------------------------------------------------------
    #                              Properties
    # -------------------------------------------------------------------------

    @property
    def type(self):
        return 'Mesh_recorder'

    @property
    def _mesh(self):
        return self._mesh_manager.mesh
        
    # -------------------------------------------------------------------------
    #                                Shooting
    # -------------------------------------------------------------------------
        
    def shoot(self, kink=False):
        self._key_register.append([])
        self.__shooting = True
        if kink:
            self._kinks = np.hstack((self._kinks, [True]))
        else:
            self._kinks = np.hstack((self._kinks, [False]))
        self._record_mesh()
        self._mesh_manager.update_plot()
        self.__shooting = False
        self._reset_interpolation()
        
    def plot_frame(self, frame_number):
        self._frame_number = frame_number
        self.__recording = True
        self._update_mesh()
        self._mesh_manager.update_plot()
        k = np.max(np.where(self._key_frames_number <= frame_number)[0]) 
        for name in self._key_register[k]:        
            if not self._sources_on_scene[name][k]:
                self._mesh_manager.hide(name)
            else:
                self._mesh_manager.show(name)
        self.__recording = False
        self._restore_mesh()
        
    def plot_key_frame(self, key_number):
        self._key_number = key_number
        self.__key = True
        if not self.__indices_updated:
            self._make_frame_register()
        self._mesh.vertices = np.copy(self._vertices[key_number])
        self._mesh_manager.selected_vertices = self._selected_vertices[key_number]
        self._mesh_manager.selected_edges = self._selected_edges[key_number]
        self._mesh_manager.selected_faces = self._selected_faces[key_number]       
        self._mesh_manager.update_plot()
        for name in self._key_register[key_number]:  
            if not self._sources_on_scene[name][key_number]:
                self._mesh_manager.hide(name)
            else:
                self._mesh_manager.show(name)
        self.__key = False
        self._restore_mesh()
         
    def process(self, source_type=None, **kwargs):
        if self.__shooting:
            self._key_register[-1].append(kwargs['name'])
            if source_type == 'faces':
                self._record_faces(**kwargs)
            elif source_type == 'edges':
                self._record_edges(**kwargs)
            elif source_type == 'vertices':
                self._record_vertices(**kwargs)
            elif source_type == 'glyph':
                self._record_glyph(**kwargs)
        elif self.__recording:
            if source_type == 'faces':
                kwargs = self._plot_faces(**kwargs)
            elif source_type == 'edges':
                kwargs = self._plot_edges(**kwargs)
            elif source_type == 'vertices':
                kwargs = self._plot_vertices(**kwargs)
            elif source_type == 'glyph':
                kwargs = self._plot_glyph(**kwargs)
        elif self.__key:
            if source_type == 'faces':
                kwargs = self._key_faces(**kwargs)
            elif source_type == 'edges':
                kwargs = self._key_edges(**kwargs)
            elif source_type == 'vertices':
                kwargs = self._key_vertices(**kwargs)
            elif source_type == 'glyph':
                kwargs = self._key_glyph(**kwargs)
        return kwargs
                
    # -------------------------------------------------------------------------
    #                                Record
    # -------------------------------------------------------------------------
    
    def clear(self):
        self._key_frames_number = None   
        self._frame_number = 0      
        self._key_number = 0        
        self._N = None        
        self._vertices = None
        self._faces = {}
        self._edges = {}
        self._points = {}
        self._selected_faces = []
        self._selected_edges = []       
        self._selected_vertices = []       
        self._glyphs = {}        
        self._interpolated_data = {}       
        self._sources_on_scene = {}       
        self._key_register = []
        self._kinks = np.array([], dtype=np.bool)
        self.__recording = False        
        self.__shooting = False
        self.__key = False 
                   
    def _record_faces(self, **kwargs):
        f1 = kwargs.get('face_data', np.zeros(self._mesh.F))        
        f0 = self._faces.get('face_data', np.empty((0,self._mesh.F)))
        self._faces['face_data'] = np.vstack((f0, f1))
        v1 = kwargs.get('vertex_data', np.zeros(self._mesh.V))
        v0 = self._faces.get('vertex_data', np.empty((0,self._mesh.V)))
        self._faces['vertex_data'] = np.vstack((v0, v1))
        
    def _record_edges(self, **kwargs):
        f1 = kwargs.get('edge_data', np.zeros(self._mesh.E)) 
        f0 = self._edges.get('edge_data', np.empty((0,self._mesh.E)))
        self._edges['edge_data'] = np.vstack((f0, f1))
        v1 = kwargs.get('vertex_data', np.zeros(self._mesh.V))
        v0 = self._edges.get('vertex_data', np.empty((0,self._mesh.V)))
        self._edges['vertex_data'] = np.vstack((v0, v1))
        
    def _record_vertices(self, **kwargs):
        v1 = kwargs.get('vertex_data', np.zeros(self._mesh.V))
        v0 = self._points.get('vertex_data', np.empty((0,self._mesh.V)))
        self._points['vertex_data'] = np.vstack((v0, v1[np.newaxis,:]))
        i1 = kwargs.get('vertex_indices', [])
        i0 = self._points.get('vertex_indices', [])
        self._points['vertex_indeices'] = i0.append(i1)
        
    def _record_glyph(self, **kwargs):
        name = kwargs['name']
        i1 = kwargs.get('vertex_indices', None)
        glyph = self._glyphs.get(name, {'vertex_indices': []})
        i0 = glyph['vertex_indices']
        i0.append(i1)
        glyph['vertex_indices'] = i0
        self._glyphs[name] = glyph
    
    def _record_mesh(self):
        if self._vertices is None:
            self._vertices = np.copy(self._mesh.vertices[np.newaxis,:])
        else:
            self._vertices = np.vstack((np.copy(self._vertices), 
                                        self._mesh.vertices[np.newaxis,:]))
        self._selected_faces.append(self._mesh_manager.selected_faces)
        self._selected_edges.append(self._mesh_manager.selected_edges)
        self._selected_vertices.append(self._mesh_manager.selected_vertices)
                                                   
    # -------------------------------------------------------------------------
    #                            Plot functions
    # -------------------------------------------------------------------------
    
    def _update_mesh(self):
        data = self._interpolated_data['vertices'][self._frame_number]
        self._mesh.vertices = np.copy(data)
        k = np.max(np.where(self._key_frames_number <= self._frame_number)[0]) 
        self._mesh_manager.selected_vertices = self._selected_vertices[k]
        self._mesh_manager.selected_edges = self._selected_edges[k]
        self._mesh_manager.selected_faces = self._selected_faces[k]
        
    def _restore_mesh(self):
        self._mesh.vertices = np.copy(self._vertices[-1]) 
        self._mesh_manager.selected_vertices = np.copy(self._selected_vertices[-1])
        self._mesh_manager.selected_edges = np.copy(self._selected_edges[-1])
        self._mesh_manager.selected_faces = np.copy(self._selected_faces[-1])
                   
    def _plot_faces(self, **kwargs):
        if 'face_data' in kwargs:
            data = self._interpolated_data['faces']['face_data']
            kwargs['face_data'] = data[self._frame_number]
        elif 'vertex_data' in kwargs:
            data = self._interpolated_data['faces']['vertex_data']
            kwargs['vertex_data'] = data[self._frame_number]
        return kwargs
    
    def _plot_edges(self, **kwargs):
        if 'edge_data' in kwargs:
            data = self._interpolated_data['edges']['edge_data']
            kwargs['edge_data'] = data[self._frame_number]
        elif 'vertex_data' in kwargs:
            data = self._interpolated_data['edges']['vertex_data']
            kwargs['vertex_data'] = data[self._frame_number]
        return kwargs
            
    def _plot_vertices(self, **kwargs):
        if 'vertex_data' in kwargs:
            data = self._interpolated_data['points']['vertex_data']
            kwargs['vertex_data'] = data[self._frame_number]
        if 'vertex_indices' in kwargs:
            data = self._interpolated_data['points']['indices']
            kwargs['vertex_indices'] = data[self._frame_number]
        return kwargs
            
    def _plot_glyph(self, **kwargs):
        if True:#'vertex_indices' in kwargs:
            data = self._interpolated_data[kwargs['name']]['vertex_indices']
            kwargs['vertex_indices'] = data[self._frame_number]
        return kwargs
    
    # -------------------------------------------------------------------------
    #                           Key Plot functions
    # -------------------------------------------------------------------------
                    
    def _key_faces(self, **kwargs):
        if 'face_data' in kwargs:
            data = self._faces['face_data']
            kwargs['face_data'] = data[self._key_number]
        elif 'vertex_data' in kwargs:
            data = self._faces['vertex_data']
            kwargs['vertex_data'] = data[self._key_number]
        return kwargs
    
    def _key_edges(self, **kwargs):
        if 'edge_data' in kwargs:
            data = self._edges['edge_data']
            kwargs['edge_data'] = data[self._key_number]
        elif 'vertex_data' in kwargs:
            data = self._edges['vertex_data']
            kwargs['vertex_data'] = data[self._key_number]
        return kwargs
            
    def _key_vertices(self, **kwargs):
        if 'vertex_data' in kwargs:
            data = self._point['vertex_data']
            kwargs['vertex_data'] = data[self._key_number]
        if 'vertex_indices' in kwargs:
            data = self._point['vertex_indices']
            kwargs['vertex_indices'] = data[self._key_number]
        return kwargs
            
    def _key_glyph(self, **kwargs):
        if 'vertex_indices' in kwargs:
            data = self._glyphs[kwargs['name']]['vertex_indices']
            kwargs['vertex_indices'] = data[self._key_number]
        return kwargs
    
    # -------------------------------------------------------------------------
    #                            Interpolation
    # -------------------------------------------------------------------------

    def interpolate_frames(self, N, key_frames_number):
        self._key_frames_number = key_frames_number
        self._N = N
        self._reset_interpolation()
        self._make_frame_register()
        self._interpolate_frames()
        
    def _reset_interpolation(self):
        self._sources = {}
        self._sources_on_scene = {}
        self._interpolated_data = {}   
        self.__indices_updated = False
            
    def _make_frame_register(self):       
        K = len(self._key_register)
        self._sources = {}
        self._sources_on_scene = {}
        for key_frame in self._key_register:
            for name in key_frame:
                self._sources[name] = 1
        for name in self._sources:
            self._sources_on_scene[name] = np.zeros(K, dtype=np.bool)
        for k in range(K):
            for name in self._key_register[k]:
                on_scene = 1
                try:
                    if name not in self._key_register[k+1]:
                        on_scene = 0
                except:
                    pass
                self._sources_on_scene[name][k] = on_scene
        self._sources_on_scene['vertices'] = np.ones(K)
        self._update_indices()
                
    def _interpolate_datum(self, source_name, data):
        on_scene = self._sources_on_scene[source_name]
        kink_1 = np.logical_or(np.roll(on_scene, 1) == 0,
                               np.roll(on_scene, -1) == 0)
        diff = np.abs(data - np.roll(data, 1))
        #thr = 0.1 * np.max(diff)
        #kink_2 = np.logical_or(diff < thr, np.roll(diff,-1) < thr)
        kink_2 = np.abs((diff - np.roll(diff,-1))/(diff + np.roll(diff,-1) + 1e-10)) > 0.05
        kink = np.logical_or(kink_1, kink_2)
        kink = np.logical_or(kink, self._kinks)
        kink[0] = kink[-1] = 1
        kink_frames = self._key_frames_number[kink]
        smooth = np.invert(np.in1d(np.arange(self._N), kink_frames))
        smooth[0] = smooth[-1] = 0
        s = np.arange(self._N)[smooth]
        S = len(s)
        K = len(self._key_frames_number)
        one = 0.1*np.ones(S)
        i0 = np.arange(K)
        i1 = np.arange(S) + K
        i = np.hstack((i0, i1, i1, i1))
        j = np.hstack((self._key_frames_number, s, s+1, s-1))
        entry = np.hstack((np.ones(K), one, -.5*one, -.5*one))
        M = sparse.coo_matrix((entry, (i,j)), shape=(S+K, self._N))
        r = np.hstack((data, np.zeros(M.shape[0] - data.shape[0])))            
        M = M.tocsc()
        H = sparse.spmatrix.dot(M.transpose(), M)
        r = sparse.spmatrix.dot(M.transpose(), r)
        X = linalg.spsolve(H, r)
        return X
    
    def _update_indices(self):
        if not self.__indices_updated:
            for key in self._glyphs:
                indices = []
                k = 0
                for i in range(len(self._key_register)):
                    if self._sources_on_scene[key][i]:
                        indices.append(self._glyphs[key]['vertex_indices'][k])
                        k += 1
                    else:
                        indices.append([])
                self._glyphs[key]['vertex_indices'] = indices
            self.__indices_updated = True
    
    def _interpolate_indices(self, name, indices):
        x = []
        k = 0
        m = 0
        for i in range(self._N):
            on_scene = False
            if self._sources_on_scene[name][k]:
                on_scene = True            
            if i >= self._key_frames_number[k]:
                k += 1
                if on_scene:
                    m += 1
            if on_scene:
                x.append(indices[m-1])
            else:
                x.append([])
        return x
    
    def _interpolate_vector(self, name, V):
        Vi = np.zeros((self._N, V.shape[1], V.shape[2]))
        for i in range(V.shape[1]):
            for j in range(V.shape[2]):
                Vi[:,i,j] = self._interpolate_datum(name, V[:,i,j])
        return Vi
        
    def _interpolate_data(self, name, D):
        Di = np.zeros((self._N, D.shape[1]))
        for i in range(D.shape[1]):
            Di[:,i] = self._interpolate_datum(name, D[:,i])
        return Di
    
    def _interpolate_frames(self):
        self._interpolated_data = {}
        Vi = self._interpolate_vector('vertices', self._vertices)
        self._interpolated_data['vertices'] = Vi
        faces = {}
        for key in self._faces:
            faces[key] = self._interpolate_data('faces', self._faces[key])
        self._interpolated_data['faces'] = faces
        edges = {}
        for key in self._edges:
            edges[key] = self._interpolate_data('edges', self._edges[key])
        self._interpolated_data['edges'] = edges
        vertices = {}
        for key in self._points:
            vertices[key] = self._interpolate_data('points', self._points[key])
        self._interpolated_data['points'] = vertices
        for key in self._glyphs:
            i = self._interpolate_indices(key, self._glyphs[key]['vertex_indices'])
            data = {'vertex_indices': i}
            self._interpolated_data[key] = data